@@ -15,7 +15,7 @@
 #include <qpdf/QPDF_Null.hh>
 #include <qpdf/QPDF_Dictionary.hh>
 
-std::string QPDF::qpdf_version = "2.2.4";
+std::string QPDF::qpdf_version = "2.3.0";
 
 void
 QPDF::InputSource::setLastOffset(off_t offset)
@@ -403,10 +403,9 @@ QPDF::parse(char const* password)
 	this->file->rewind();
     }
     char* buf = new char[tbuf_size + 1];
-    // Put buf in a PointerHolder to guarantee deletion of buf.  This
-    // calls delete rather than delete [], but it's okay since buf is
-    // an array of fundamental types.
-    PointerHolder<char> b(buf);
+    // Put buf in an array-style PointerHolder to guarantee deletion
+    // of buf.
+    PointerHolder<char> b(true, buf);
     memset(buf, '\0', tbuf_size + 1);
     this->file->read(buf, tbuf_size);
 
@@ -1887,6 +1886,39 @@ QPDF::getObjectByID(int objid, int generation)
     return QPDFObjectHandle::Factory::newIndirect(this, objid, generation);
 }
 
+void
+QPDF::replaceObject(int objid, int generation, QPDFObjectHandle oh)
+{
+    if (oh.isIndirect())
+    {
+	QTC::TC("qpdf", "QPDF replaceObject called with indirect object");
+	throw std::logic_error(
+	    "QPDF::replaceObject called with indirect object handle");
+    }
+
+    // Force new object to appear in the cache
+    resolve(objid, generation);
+
+    // Replace the object in the object cache
+    ObjGen og(objid, generation);
+    this->obj_cache[og] =
+	ObjCache(QPDFObjectHandle::ObjAccessor::getObject(oh), -1, -1);
+}
+
+void
+QPDF::swapObjects(int objid1, int generation1, int objid2, int generation2)
+{
+    // Force objects to be loaded into cache; then swap them in the
+    // cache.
+    resolve(objid1, generation1);
+    resolve(objid2, generation2);
+    ObjGen og1(objid1, generation1);
+    ObjGen og2(objid2, generation2);
+    ObjCache t = this->obj_cache[og1];
+    this->obj_cache[og1] = this->obj_cache[og2];
+    this->obj_cache[og2] = t;
+}
+
 void
 QPDF::trimTrailerForWrite()
 {
